#pragma once

extern "C" {

	/**
	* Print RadISH banner
	*/
	void print_banner();

	/**
	* Initialise RadISH.
	*/
	void initialise_radish(int number_of_events, const char* pdf_set, int pdf_set_length, int pdf_member, const char* collider, int collider_lenght, const char* process, int process_length, double sqrts, double Q_over_M, double modified_log_exponent, double jet_radius, double jet_veto, int resummation_order, const char* observable, int observable_length, int random_gen_seed);

	/**
	* Book n_scales histograms for a given observable, with n_bins bins and edges bin_edges
	*/
	void book_matrix_histogram(const char* histogram_name, int histogram_name_length, const char* observable, int observable_length, int n_scales, double bin_edges[], int n_bins);

	/**
	* Book an histogram for a given observable with a given bin_index and a given scale_index, with n_bins bins and edges bin_edges
	*/
	void book_geneva_histogram(const char* observable, int observable_length, int bin_index, int scale_index, double bin_edges[], int n_bins);

	/**
	* Start RadISH shower for one phase-space-point, filling the histograms for all the observables given a scale index.
	*/
	void gen_rad_matrix(double* born_momenta_array, int* born_particles_id, int n_particles, double *msqB, double *msqV1, double *msqV2, int nf, double* jacobian_in, double *shower_wgt, int scale_index);

	/**
	* Start RadISH shower for one phase-space-point, filling the histograms for all the observables given a scale index.
	*/
	void gen_rad_geneva(double* pA_in, int idA_in, double* pB_in, int idB_in, double msqB, double msqV1, double msqV2, int nf_in, double* jacobian_in, int scale_index, int bin_index_in, int index);

	/**
	* Terminate RadISH and clear the memory
	*/
	void terminate_radish();

	/**
	* Print all the histograms to file. Number of Born events passing the cuts are needed for normalization.
	*/
	void print_all_histos(int number_of_Born_events);

	/**
	* Return histogram once it is filled and its norm
	*/
	void get_geneva_histo(int index, int nbins, double* pt, double* res, double *reserr, double* exp, double* exp_as, int fixed_order, double* norm);

	/**
	* Init default EW parameters
	*/
	void init_default();

	/**
	* Whether to use numerical routine for jetveto (ptj) resummation
	*/
	void use_numerical_jetveto();

	/**
	* Set dynamic scales
	*/
	void update_dynamical_parameters(double *muR, double *muF, double *Msinglet);

    /**
	* Set resummation scale ratio
	*/
	void set_q_over_m(double *Q_over_M);

	/**
	* Use prime counting
	*/
	void set_prime_counting();

    /**
	* Set the scale index to i
	*/
	void change_index(int i);

	/**
	* Print one sigle histogram
	*/
	void print_one_histo(int i, int n_born, int n_events);

    /**
	* Init Yukawa couplings
	*/
	void init_yukawas(double *ybottom, double *ycharm, double *ystrange);

	/**
	* Init custom EW parameters
	*/
	void init_external(double *MZin, double *MWin, double * Wwidthin,double * Zwidthin, double * Hwidthin,double * aemmzin,double * sinwsqin,double * higgs_vevin,double * msin,double * mcin, double * mbin, double * mtin, double * mHin);

	/**
	* Matching
	*/
	void do_matching(const char *scheme, int nbins, int fixed_order, int resum_order, double* pt, double *sigma, double *sigma_err, double *lumi_asympt, double *lumi_asympt_err, double *sigmabar_1, double *sigmabar_1_err, double *sigmabar_2, double *sigmabar_2_err, double *sigmabar_3, double *sigmabar_3_err, double *resummed, double *resummed_err, double *res_expanded_1, double *res_expanded_1_err, double *res_expanded_2, double *res_expanded_2_err, double *res_expanded_3, double *res_expanded_3_err, double *matched_cumulative, double *fixed_order_result, bool normalised, bool damping, const char *observable, double scale);

	void do_matching_matrix(const char *scheme, int nbins, int fixed_order, int resum_order, double *sigmatot, double *sigma_err, double *lumi_asympt, double *lumi_asympt_err, double *sigma_1, double *sigma_1_err, double *sigma_2, double *sigma_2_err, double *sigma_3, double *sigma_3_err, double *resummed, double *resummed_err, double *res_expanded_1, double *res_expanded_1_err, double *res_expanded_2, double *res_expanded_2_err, double *res_expanded_3, double *res_expanded_3_err, double *matched_cumulative, double *fixed_order_result, bool normalised);

}
